#include "gd.h"
#include "gdfontg.h"

#include "gdstuff.h"


/// Create a jpeg in memory, containing the supplied text in white on a black background.
/// header:		Input:  The header text to be rendered in a title bar at the top of the jpeg image. Max 51 characters.
/// text1:		Input:  The first line of text to be rendered near the middle of the jpeg image. Max 51 characters.
/// text2:		Input:  The second line of text to be rendered near the middle of the jpeg image. Max 51 characters.
/// footer:		Input:  The footer text to be rendered at the bottom of the jpeg image. Max 51 characters.
/// size:		Output: A pointer to an integer that will receive the size of the returned JPEG image memory block.
/// Returns:	A pointer to the JPEG image in memory.  NOTE: This memory must be freed by calling free_gd_mem(). 
void * create_text_jpeg(
	const unsigned char* header, 
	const unsigned char* text1, 
	const unsigned char* text2, 
	const unsigned char* footer, int *size)
{
	const int imgWidth  = 512;
	const int imgHeight = 288;

	gdFontPtr font = gdFontGetGiant();
	gdImagePtr img = gdImageCreateTrueColor(imgWidth, imgHeight);
	int blue	= gdImageColorAllocate(img,  32,  32, 128);
	int white	= gdImageColorAllocate(img, 255, 255, 255);
	int grey	= gdImageColorAllocate(img, 128, 128, 128);

	gdImageFilledRectangle (img, 0,  0, imgWidth, imgHeight, blue);
	gdImageFilledRectangle (img, 0, 20, imgWidth, 54, grey);

	if (header)
		gdImageString (img, font,  32,  30, (unsigned char *)header, blue);

	if (text1)
		gdImageString (img, font,  32, 124, (unsigned char *)text1, white);

	if (text2)
		gdImageString (img, font,  32, 144, (unsigned char *)text2, white);

	if (footer)
		gdImageString (img, font,  32, 226, (unsigned char *)footer, grey);

	void * retval = gdImageJpegPtr(img, size, 100);

	gdImageDestroy(img);

	return retval;
}

void free_gd_mem(void * mem)
{
	gdFree(mem);
}
